/*
 * Copyright (c) 2018
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: child2.c 1390 2018-11-16 17:36:02Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "acld.h"
#include "child2.h"

/* Return 1 if we did anything */
void
child2connect(struct state *sp, fd_set *rp, fd_set *wp)
{
	struct child *chp;

	chp = sp->chp2;
// lg(LOG_DEBUG, "# %s: child2connect: entry", chp->name);
	if (chp->state == ASTATE_CONNECTING &&
	    chp->f_childconnect != NULL &&
	    ((chp->rfd >= 0 && FD_ISSET(chp->rfd, rp)) ||
	    (chp->wfd >= 0 && FD_ISSET(chp->wfd, wp))))
		(chp->f_childconnect)(chp);
}

void
child2filter(struct state *sp, struct client *cl, struct req *rp)
{
	struct child *chp;

	chp = sp->chp2;
	if (chp->f_childfilter != NULL) {
		(chp->f_childfilter)(chp, cl, rp);
		
	}
}


void
child2input(struct state *sp)
{
	struct child *chp;
	struct req *rp;

	chp = sp->chp2;
	if (chp->f_childinput != NULL) {
		while (IOBUF_LEN(&chp->rbuf) > 0 && !(chp->f_childinput)(chp))
			continue;
		rp = chp->req;
		if (rp != NULL && rp->state == RSTATE_DONE) {
			/* Remove from request list */
			chp->req = rp->next;
			rp->next = NULL;
			freereq(sp, rp);
			return;
		}
	}
	return;
}

void
child2poll(struct state *sp, fd_set *rp, fd_set *wp)
{
	ssize_t cc;
	struct child *chp;

	chp = sp->chp2;
	switch (chp->state) {

	case ASTATE_NOTCONNECTED:
		/* Reset timers if not pending */
		if (chp->f_childlogin != NULL) {
			if (timercheck(&chp->t_login) < 0) {
				timerset(&chp->t_login, cf->c_login_secs2);
			}
		}

		/* Did the child fall asleep? */
		if (chp->f_childlogin != NULL && timerdue(&chp->t_login)) {
			(chp->f_childlogin)(chp);
			return;
		}
		break;

	case ASTATE_CONNECTING:
		/* Don't interfer with connect */
		return;

	case ASTATE_CONNECTED:
		if (chp->f_childayt != NULL) {
			(chp->f_childayt)(chp);
			child2setastate(chp, ASTATE_SENTLOGIN);
			return;
		}
		break;

	case ASTATE_LOGGEDIN:
		/* Poke the child */
		if (chp->f_childayt != NULL && IOBUF_LEN(&chp->wbuf) == 0 &&
		    timerdue(&chp->t_ayt))
			(chp->f_childayt)(chp);
		break;

	default:
		break;
	}

#ifdef notdef
	if (chp->f_childkill != NULL && chp->wfd >= 0 &&
	    chp->state == ASTATE_LOGGEDIN)
		(chp->f_childkill)(chp);
#endif

#ifdef notdef
	/* Don't interfer with connect */
	if (chp->state == ASTATE_CONNECTING)
		return;
#endif

	/* Read */
	if (chp->f_childread != NULL &&
	    chp->rfd >= 0 && FD_ISSET(chp->rfd, rp)) {
		cc = (chp->f_childread)(chp);
		if (cc <= 0) {
			if (cc == 0)
				lg(LOG_DEBUG, "%s: child disconnect",
				    chp->name);
			else
				lg(LOG_ERR, "%s: child read: %s",
				    chp->name, strerror(errno));
			if (chp->state != ASTATE_NOTCONNECTED)
				(chp->f_childkill)(chp);
			return;
		}
	}

	/* Write */
	if (chp->f_childwrite != NULL &&
	    chp->wfd >= 0 && FD_ISSET(chp->wfd, wp)) {
		cc = (chp->f_childwrite)(chp);
		if (cc <= 0) {
			if (cc == 0)
				lg(LOG_DEBUG, "%s: child disconnect",
				    chp->name);
			else
				lg(LOG_ERR, "%s: second child write: %s",
				    chp->name, strerror(errno));
			if (chp->state != ASTATE_NOTCONNECTED)
				(chp->f_childkill)(chp);
			return;
		}
	}
	return;
}

void
child2setastate(struct child *chp, enum acldstate astate2)
{
	/* Track logged in vs. not logged in */
// lg(LOG_DEBUG, "# astate2 %s", val2str(astate2str, astate2));
	if ((chp->state == ASTATE_LOGGEDIN) ^ (astate2 == ASTATE_LOGGEDIN))
		chp->loginout_ts = time(NULL);
	chp->state = astate2;
}

/* Return select seconds or -1 */
int
child2settimeout(struct state *sp)
{
	int t, secs;
	struct child *chp;

	secs = -1;
	chp = sp->chp2;
	if (chp->f_childlogin != NULL) {
		t = timercheck(&chp->t_login);
		if (t >= 0 && secs > t)
			secs = t;
	}
	if (chp->f_childayt != NULL) {
		t = timercheck(&chp->t_ayt);
		if (t >= 0 && secs > t)
			secs = t;
	}
	return (secs);
}

void
child2state(struct state *sp, struct req *rp)
{
	const char *p;
	struct child *chp;
	char buf[256];

	chp = sp->chp2;
	ioappendfmt(&rp->payload, "# %s", chp->name);
	p = val2str(astate2str, chp->state);
	if (p == NULL) {
		(void)snprintf(buf, sizeof(buf), "#%d", chp->state);
		p = buf;
	}
	ioappendfmt(&rp->payload, "# state2 %s", p);
}

void
child2reload(struct state *sp)
{
	struct child *chp;

	chp = sp->chp2;
	if (chp->f_childkill != NULL && chp->wfd >= 0 &&
	    chp->state == ASTATE_LOGGEDIN)
		(chp->f_childkill)(chp);
}

int
child2update_fd_set(struct state *sp, fd_set *rp, fd_set *wp, int nfds)
{
	struct child *chp;

	chp = sp->chp2;
	if (chp->rfd >= 0) {
		FD_SET(chp->rfd, rp);
		nfds = MAX(nfds, chp->rfd);
	}

	if (chp->wfd >= 0) {
		FD_SET(chp->wfd, wp);
		nfds = MAX(nfds, chp->wfd);
	}
	return (nfds);
}
