/*
 * Copyright (c) 2018
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: corsa.c 1415 2018-11-21 06:22:56Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>

#include <netinet/tcp.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include <json-c/json.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/x509_vfy.h>

#include "acld.h"
#include "child2.h"
#include "corsa.h"
#include "sslutil.h"
#include "timer.h"

/* Locals */
static const char gigafilter_uri[] = "/api/v1/gigafilter";
static const char http11[] = "HTTP/1.1";

/* XXX move to cf? */
static u_short port = 443;

/* Forwards */
static void corsaadd(struct child *, struct client *, struct req *);
static void corsaayt(struct child *);
static void corsakill(struct child *);
static void corsalogin(struct child *);
static void corsaparseuptime(struct child *, json_object *);
static void corsafilter(struct child *, struct client *, struct req *);
static int corsareadwrite(struct child *);
static void corsasend(struct child *, const char *, const char *,
    const char *);

static void corsaconnect(struct child *);

static void corsaremove(struct child *, struct client *, struct req *);
static void corsaquery(struct child *, struct client *, struct req *);

static int corsainput(struct child *);
static int corsaparse(struct child *, struct client *, struct req *,
    const char *, int);

static int goodfilter(struct addr *);

int
corsaaddnet(const char *str)
{
	const char *errmsg;
	struct addr *a;
	struct array *dp;

	dp = &cf->c_corsalocalnets_array;
	DYNARRAY(dp, cf->c_corsalocalnets, dp->len + 1, "corsa nets");
	a = cf->c_corsalocalnets + dp->len;
	errmsg = extractaddr(str, NULL, a);
	if (errmsg != NULL) {
		lg(LOG_ERR, "corsaaddnet: extractaddr: %s", errmsg);
		return (0);
	}

	if (a->family != AF_INET) {
		lg(LOG_ERR, "corsaaddnet: Not an IPv4 network: %s", str);
		return (0);
	}

	++dp->len;
	return (1);
}

static void
corsaayt(struct child *chp)
{
	struct req *rp;

	rp = newreq("corsaserverayt req");
	rp->state = RSTATE_PENDING;
	rp->type = REQ_AYT;
	rp->cmd = "ayt";

	/* Arrival timestamp */
	getts(&rp->ats);

	/* Append the new request to end of the list */
	appendreq(&chp->req, rp);

	/* Add the request to the iobuf */
	corsasend(chp, "/api/v1/equipment/cpu", "GET", NULL);
}

static void
corsaconnect(struct child *chp)
{
	int r, err;
	struct sslclient *sup;

	sup = (struct sslclient *)chp->opaque;
	if (chp->rfd >= 0) {
		r = SSL_connect(sup->ssl);
		if (r == 0) {
			err = SSL_get_error(sup->ssl, r);
			lg(LOG_DEBUG, "%s: connection to %s failed: %s",
			    chp->name, ERR_reason_error_string(err),
			    sup->remote);
			child2setastate(chp, ASTATE_NOTCONNECTED);
			timerset(&chp->t_login, 1);
			return;
		}

		if (r < 0) {
			if (!sslhandleerr(sup, r, "corsaconnect"))
				corsakill(chp);
			return;
		}
	}

	/* We're connected now */
	lg(LOG_DEBUG, "%s: connected to %s", chp->name, sup->remote);
	child2setastate(chp, ASTATE_CONNECTED);
	chp->numsends = 0;

	/* We only need to select() on read */
	chp->rfd = sup->s;
	chp->wfd = -1;
}

static void
corsafilter(struct child *chp, struct client *cl, struct req *rp)
{
	struct addr *wp;
	char buf[64];

	switch (rp->type) {

	case REQ_FILTER:
	case REQ_NOFILTER:
	case REQ_QUERYFILTER:
		break;

	default:
		lg(LOG_ERR, "%s: corsafilter: bad type %d",
		    chp->name, rp->type);
		exit(EX_SOFTWARE);
	}

	/* We can only handle IPv4 */
	if (rp->filter.family != AF_INET) {
		clientsenderr(cl, rp, "%s",
		    "Only IPv4 addresses are supported");
		return;
	}

	/* Check for required corsalocalnets cidr list */
	if (cf->c_corsalocalnets == NULL) {
		clientsenderr(cl, rp, "%s",
		    "corsalocalnets missing from configuration");
		return;
	}

	/* Check that the address is not a local filter network */
	if (!goodfilter(&rp->filter)) {
		clientsenderr(cl, rp, "%s part of a configured local filternet",
		    addr2str(&rp->filter));
		return;
	}

	/* Bail if we're not ready to handle general requests yet */
	if (chp->state != ASTATE_LOGGEDIN)
		return;

	switch (rp->type) {

	case REQ_FILTER:
		wp = whitelist(&rp->filter);
		if (wp != NULL) {
			snprintf(buf, sizeof(buf), "%s", addr2str(&rp->filter));
			clientsenderr(cl, rp, "%s is on the whitelist (%s)",
			    buf, addr2str(wp));

			/* Log for NETS */
			nets_log(cl, rp);
			return;
		}
		stats_setrate(&chp->sp->filterstats);
		corsaadd(chp, cl, rp);
		break;

	case REQ_NOFILTER:
		corsaremove(chp, cl, rp);
		break;

	case REQ_QUERYFILTER:
		corsaquery(chp, cl, rp);
		break;

	default:
		lg(LOG_ERR, "%s: corsafilter: can't happen (%d)",
		    chp->name, rp->type);
		exit(EX_SOFTWARE);
	}

	/* Update request state */
	rp->state = RSTATE_CHILD;
}

void
corsainit(struct state *sp)
{
	int i;
	struct child *chp;
	struct sslclient *sup;
	char buf[64];

	/* Static initialization */
	chp = new(sizeof(*chp), 1, "corsa child");
	i = 1;
	snprintf(buf, sizeof(buf), "corsa%d", i);
	chp->name = strsave(buf);
	chp->sp = sp;

	/* Static initialization */
	IOBUF_INIT(&chp->rbuf);
	IOBUF_INIT(&chp->wbuf);

	sp->chp2 = chp;

	/* SSL */
	sup = new(sizeof(*sup), 1, "corsa SSL");
	chp->opaque = (void *)sup;
	sup->name = chp->name;
	sup->rfdp = &chp->rfd;
	sup->wfdp = &chp->wfd;
	sup->doingwrite = 0;

	/* Openssl */
	SSL_load_error_strings();
	(void)SSL_library_init();

	chp->rfd = -1;
	chp->wfd = -1;

	/* Child two hooks */
	chp->f_childayt = corsaayt;
	chp->f_childconnect = corsaconnect;
	chp->f_childfilter = corsafilter;
	chp->f_childinput = corsainput;
	chp->f_childkill = corsakill;
	chp->f_childlogin = corsalogin;
	chp->f_childread = corsareadwrite;
	chp->f_childwrite = corsareadwrite;

	stats_init(&sp->filterstats, 10, 60, "corsainit: stats");

	/* Client hooks */
	sp->f_clientfilter = child2filter;

	/* Set login timer for first attempt */
	child2setastate(chp, ASTATE_NOTCONNECTED);
	timerset(&chp->t_login, 1);
}

/* Process input from the second child, return 1 if done */

static int
corsainput(struct child *chp)
{
	long v, ret;
	char *cp, *ep;
	const char *p;
	struct iobuf *ip;
	struct client *cl;
	struct req *rp, *rp2;
	/* XXX */
	struct state *sp;

	/* Find server or client request */
	sp = chp->sp;
	rp = chp->req;
	cl = NULL;
	ip = &chp->rbuf;
	if (rp == NULL) {
		rp = NULL;
		for (cl = sp->clients; cl != NULL; cl = cl->next) {
			for (rp2 = cl->req; rp2 != NULL; rp2 = rp2->next) {
				if (rp2->state == RSTATE_CHILD) {
					rp = rp2;
				}
			}
			if (rp != NULL)
				break;
		}
	}

	if (rp == NULL) {
		lg(LOG_DEBUG, "%s: corsainput: input without request:",
		    chp->name);
		while ((cp = iogetstr(ip)) != NULL)
			if (debug > 1)
				lg(LOG_DEBUG, "corsainput: eat: \"%s\"",
				    pretty(cp));
		return (1);
	}

	/* Parse the response headers */
	ret = -1;
	while ((cp = iogetstr(ip)) != NULL) {
		if (*cp == '\r' && strcmp(cp, "\r\n") == 0)
			break;
		if (strncasecmp(cp, http11, sizeof(http11) - 1) == 0) {
			if (debug > 1)
				lg(LOG_DEBUG, "corsainput: eat: \"%s\"",
				    pretty(cp));
			cp = strchr(cp, ' ');
			if (cp != NULL) {
				++cp;
				v = strtol(cp, &ep, 10);
				if (*ep == ' ')
					ret = v;
			}
			continue;
		}
		if (debug > 1)
			lg(LOG_DEBUG, "corsainput: eat \"%s\"", pretty(cp));
	}

	if (IOBUF_LEN(ip) == 0) {
		p = "corsainput: missing json payload";
		if (cl != NULL)
			clientsenderr(cl, rp, "%s", p);
		else {
			rp->state = RSTATE_DONE;
			getts(&rp->cts);
		}
		return (1);
	}
	cp = iogetstr(ip);
	return (corsaparse(chp, cl, rp, cp, ret));
}

static void
corsakill(struct child *chp)
{
	struct iobuf *ip;
	struct sslclient *sup;
	int delta;

	if (chp->state == ASTATE_LOGGEDIN) {
		delta = time(NULL) - chp->loginout_ts;
		lg(LOG_DEBUG,
		    "%s: corsakill: logged in for %d secs (%d requests sent)",
		    chp->name, delta, chp->numsends);
	}

	sup = (struct sslclient *)chp->opaque;
	if (sup->ssl != NULL) {
		(void)SSL_shutdown(sup->ssl);
		/* Also frees BIO */
		SSL_free(sup->ssl);
		sup->ssl = NULL;
		SSL_CTX_free(sup->ssl_ctx);
		sup->ssl_ctx = NULL;
	}
	if (chp->rfd >= 0)
		(void)close(chp->rfd);
	if (chp->rfd != chp->wfd)
		(void)close(chp->wfd);
	chp->rfd = -1;
	chp->wfd = -1;
	sup->s = -1;

	ip = &chp->rbuf;
	if (IOBUF_LEN(ip) != 0)
		iofree(ip);
	IOBUF_INIT(ip);
	ip = &chp->wbuf;
	if (IOBUF_LEN(ip) != 0)
		iofree(ip);
	IOBUF_INIT(ip);
	child2setastate(chp, ASTATE_NOTCONNECTED);
	timerset(&chp->t_login, 1);

	timerreset(&chp->t_ayt);
}

static void
corsalogin(struct child *chp)
{
	const char *errmsg;
	struct addr *a;
	struct addr addr;
	struct sockaddr_storage sa;
	struct sockaddr *sap;
	struct iobuf iobuf2;
	struct iobuf *ip2;
	BIO *bio;
	socklen_t salen;
	struct sslclient *sup;

	sup = (struct sslclient *)chp->opaque;
	if (chp->rfd >= 0) {
		(void)close(chp->rfd);
		if (chp->rfd == chp->wfd)
			chp->wfd = -1;
		chp->rfd = -1;
	}
	if (chp->wfd >= 0) {
		(void)close(chp->wfd);
		chp->wfd = -1;
	}
	sup->s = -1;

	a = &addr;
	memset(a, 0, sizeof(*a));
	/* Clobber remote server */
	strlcpy(sup->remote, "?", sizeof(sup->remote));

	ip2 = &iobuf2;
	memset(ip2, 0, sizeof(*ip2));
	IOBUF_INIT(ip2);

	errmsg = extractaddr(cf->c_corsaserver, NULL, a);
	if (errmsg != NULL) {
		lg(LOG_ERR, "%s: corsainit: %s: %s",
		    chp->name, cf->c_corsaserver, errmsg);
		exit(EX_SOFTWARE);
	}

	sap = (struct sockaddr *)&sa;
	salen = addr2sa(a, port, sap);
	(void)snprintf(sup->remote, sizeof(sup->remote), "%s.%d",
	    addr2str(a), (int)port);

	sup->s = socket(a->family, SOCK_STREAM, 0);
	if (sup->s < 0) {
		lg(LOG_ERR, "%s: corsainit: %s: %s",
		    chp->name, sup->remote, strerror(errno));
		exit(EX_SOFTWARE);
	}

	/* Non-blocking */
	blocking(sup->s, 0);
	errno = 0;
	if (connect(sup->s, sap, salen) < 0) {
		if (errno != EINPROGRESS) {
			lg(LOG_ERR, "%s: connect to %s: %s",
			    chp->name, sup->remote, strerror(errno));

			/* Clean up */
			(void)close(sup->s);
			return;
		}
		lg(LOG_DEBUG, "%s: connecting to %s", chp->name, sup->remote);
	}

	if (sup->ssl != NULL) {
		(void)SSL_shutdown(sup->ssl);
		/* Also frees BIO */
		SSL_free(sup->ssl);
		sup->ssl = NULL;
	}
	if (sup->ssl_ctx != NULL) {
		SSL_CTX_free(sup->ssl_ctx);
		sup->ssl_ctx = NULL;
	}
	sup->ssl_ctx = SSL_CTX_new(SSLv23_client_method());
	if (sup->ssl_ctx == NULL) {
		lg(LOG_ERR, "%s: corsalogin: SSL_CTX_new failed: %s",
		    chp->name, ERR_error_string(ERR_get_error(), NULL));
		exit(EX_SOFTWARE);
	}
	(void)SSL_CTX_set_options(sup->ssl_ctx,
	    SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3);

	sup->ssl = SSL_new(sup->ssl_ctx);
	bio = BIO_new(BIO_s_mem());
	(void)BIO_set_nbio(bio, 1);
	SSL_set_bio(sup->ssl, bio, bio);

	SSL_set_verify(sup->ssl,
	    SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);

	if (!SSL_CTX_load_verify_locations(sup->ssl_ctx, cf->c_corsacacert,
	    NULL)) {
		lg(LOG_ERR, "%s: corsalogin: "
		    "SSL_CTX_load_verify_locations failed: %s: %s",
		    chp->name, cf->c_corsacacert, strerror(errno));
		exit(EX_SOFTWARE);
	}

	SSL_set_fd(sup->ssl, sup->s);

	chp->rfd = sup->s;

	/* Try to connect (will likely fail) */
	child2setastate(chp, ASTATE_CONNECTING);
	corsaconnect(chp);

	/* Set login retry in case this attempt fails */
	timerset(&chp->t_login, cf->c_login_secs2);
}

static int
corsaparse(struct child *chp, struct client *cl, struct req *rp,
    const char *jstr, int ret)
{
	int present;
	const char *what, *info;
	json_object *reqo, *valop;

	reqo = json_tokener_parse(jstr);
	if (reqo == NULL) {
		lg(LOG_ERR, "%s: Cannot parse json \"%s\"",
		    chp->name, pretty(jstr));
		if (cl != NULL)
			clientsenderr(cl, rp, "%s",
			    "Cannot parse json from server");
		return (1);
	}
	switch (rp->type) {

	case REQ_AYT:
		corsaparseuptime(chp, reqo);
		json_object_put(reqo);
		rp->state = RSTATE_DONE;
		getts(&rp->cts);
		return (1);

	case REQ_FILTER:
		if (ret != 200) {
			if (cl != NULL)
				clientsenderr(cl, rp, "unknown error %d", ret);
			json_object_put(reqo);
			return (1);
		}

		what = "info";
		if (json_object_object_get_ex(reqo, what, &valop)) {
			info = json_object_get_string(valop);
			if (strstr(info, "already") != 0) {
				rp->flags |= RFLAG_IGNORE;
				if (cl != NULL)
					clientsendfmt(cl, rp,
					    "Note: %s is already filtered",
					    addr2str(&rp->filter));
				break;
			}
		}
		if (cl != NULL)
			clientsend(cl, rp);
		break;

	case REQ_NOFILTER:
		if (ret != 200) {
			if (cl != NULL)
				clientsenderr(cl, rp, "unknown error %d", ret);
			json_object_put(reqo);
			return (1);
		}

		what = "info";
		if (json_object_object_get_ex(reqo, what, &valop)) {
			info = json_object_get_string(valop);
			if (strstr(info, "does not exist") != 0) {
				clientsenderr(cl, rp, "%s not found",
				    addr2str(&rp->filter));
				json_object_put(reqo);
				return (1);
			}
		}
		if (cl != NULL)
			clientsend(cl, rp);
		break;

	case REQ_QUERYFILTER:
		if (ret == 404) {
			if (cl != NULL)
				clientsenderr(cl, rp, "%s not found",
				    addr2str(&rp->filter));
			json_object_put(reqo);
			return (1);
		}

		what = "present";
		if (json_object_object_get_ex(reqo, what, &valop)) {
			present = json_object_get_int64(valop);
			if (!present) {
				if (cl != NULL)
					clientsenderr(cl, rp, "%s not present?",
					    addr2str(&rp->filter));
				json_object_put(reqo);
				return (1);
			}
		}
		if (cl != NULL)
			clientsendfmt(cl, rp, "%s",
			    addr2str(&rp->filter));
		json_object_put(reqo);
		return (1);

	default:
		lg(LOG_ERR, "%s: corsaparse: bad type %d", chp->name, rp->type);
		exit(EX_SOFTWARE);
	}

	/* Log for NETS */
	/* XXX include child name somehow? */
	if (cl != NULL)
		nets_log(cl, rp);
	json_object_put(reqo);
	return (1);
}

static void
corsaparseuptime(struct child *chp, json_object *reqo)
{
	int i;
	const char *p, *what;
	json_object *valop2;
	double avenrun[3];

	memset(avenrun, 0, sizeof(avenrun));
	json_object_object_foreach(reqo, key, valop) {
		p = strchr(key, ' ');
		if (p == NULL)
			continue;
		++p;
		switch (atoi(p)) {

		case 1:
			i = 0;
			break;

		case 5:
			i = 1;
			break;

		case 15:
			i = 2;
			break;

		default:
			continue;
		}
		what = "load-avg";
		if (i >= 0 && json_object_object_get_ex(valop, what, &valop2))
			avenrun[i] = json_object_get_double(valop2);
	}
	lg(LOG_INFO, "%s: load averages: %.2f, %.2f, %.2f",
	    chp->name, avenrun[0], avenrun[1], avenrun[2]);
}

/*
 * Read or write from a socket and update the corresponding iobuf
 *
 * Return 1 on success, -1 when the caller should report the errno.
 * Return 0 on EOF and any other other error fatal to the child.
 */
static int
corsareadwrite(struct child *chp)
{
	int r;
	struct iobuf *ip;
	struct sslclient *sup;

	sup = (struct sslclient *)chp->opaque;
	if (sup->doingwrite) {
		ip = &chp->wbuf;
		r = sslwrite(sup->ssl, ip);
		if (r > 0) {
			if (IOBUF_LEN(ip) == 0) {
				chp->rfd = sup->s;
				chp->wfd = -1;
				sup->doingwrite = 0;
			}
			return (1);
		}
	} else {
		r = sslread(sup->ssl, &chp->rbuf);
		if (r > 0) {
			if (chp->state == ASTATE_SENTLOGIN) {
				lg(LOG_DEBUG, "%s: logged in to %s",
				    chp->name, sup->remote);
				child2setastate(chp, ASTATE_LOGGEDIN);
				timerreset(&chp->t_login);
			}
			timerset(&chp->t_ayt, cf->c_ayt_secs2);
			return (1);
		}
	}

	if (!sslhandleerr(sup, r, "corsareadwrite")) {
		corsakill(chp);
		return (0);
	}
	return (1);
}

static void
corsasend(struct child *chp, const char *uri, const char *method,
    const char *jstr)
{
	char *cp, *token;
	struct iobuf *ip;
	struct sslclient *sup;

	/* corsareadtoken() logs any error */
	token = corsareadtoken(cf->c_corsatoken);
	if (token == NULL)
		return;

	sup = (struct sslclient *)chp->opaque;
	ip = &chp->wbuf;

	ioappendfmt(ip, "%s %s HTTP/1.1", method, uri);
	/* Use ip address in place of http server name */
	ioappendfmt(ip, "Host: %s", cf->c_corsaserver);
	// ioappendline(ip, "Accept-Encoding: identity");
	// ioappendfmt(ip, "User-Agent: %s", version);
	ioappendfmt(ip, "Authorization: %s", token);
	free(token);

	/* Do we have json to send? */
	if (jstr != NULL) {
		ioappendline(ip, "Content-Type: application/json");
		ioappendfmt(ip, "Content-Length: %lu", strlen(jstr));
	}

	/* Blank line */
	ioappendline(ip, "");

	if (jstr != NULL) {
		ioappendline(ip, jstr);
		/* Remove trailing /r/n */
		IOBUF_LEN(ip) -= 2;
		cp = ip->buf + IOBUF_LEN(ip);
		*cp = '\0';
	}
	sup->doingwrite = 1;
	chp->rfd = -1;
	chp->wfd = sup->s;
	++chp->numsends;
}

static void
corsaadd(struct child *chp, struct client *cl, struct req *rp)
{
	const char *jstr;
	json_object *reqo, *itemop;
	struct addr *a;
	char path[128];

	a = &rp->filter;
	(void)snprintf(path, sizeof(path), "/ipv4/%s", addr2str(a));
	itemop = json_object_new_object();
	json_object_object_add(itemop, "op", json_object_new_string("add"));
	json_object_object_add(itemop, "path", json_object_new_string(path));
	json_object_object_add(itemop, "value", NULL);

	reqo = json_object_new_array();
	json_object_array_add(reqo, itemop);

	jstr = json_object_to_json_string(reqo);

	corsasend(chp, gigafilter_uri, "PATCH", jstr);

	json_object_put(reqo);
}

static void
corsaremove(struct child *chp, struct client *cl, struct req *rp)
{
	const char *jstr;
	json_object *reqo, *itemop;
	char path[128];
	struct addr *a;

	a = &rp->filter;
	(void)snprintf(path, sizeof(path), "/ipv4/%s", addr2str(a));
	itemop = json_object_new_object();
	json_object_object_add(itemop, "op", json_object_new_string("remove"));
	json_object_object_add(itemop, "path", json_object_new_string(path));

	reqo = json_object_new_array();
	json_object_array_add(reqo, itemop);

	jstr = json_object_to_json_string(reqo);

	corsasend(chp, gigafilter_uri, "PATCH", jstr);

	json_object_put(reqo);
}

static void
corsaquery(struct child *chp, struct client *cl, struct req *rp)
{
	char uri[128];
	struct addr *a;

	a = &rp->filter;
	(void)snprintf(uri, sizeof(uri), "%s?ipv4=%s",
	    gigafilter_uri, addr2str(a));

	corsasend(chp, uri, "GET", NULL);
}

/* Returns a malloc() buffer with the token string or else NULL */
char *
corsareadtoken(const char *fn)
{
	size_t n;
	FILE *f;
	char *buf;
	char *cp;
	struct stat sbuf;

	f = fopen(fn, "r");
	if (f == NULL) {
		lg(LOG_ERR, "corsareadtoken: fopen %s: %s",
		    fn, strerror(errno));
		return (NULL);
	}
	if (fstat(fileno(f), &sbuf) < 0) {
		lg(LOG_ERR, "corsareadtoken: fstat %s: %s",
		    fn, strerror(errno));
		(void)fclose(f);
		return (NULL);
	}
	n = sbuf.st_size;
	buf = new(n + 1, 1, "corsa auth token");
	if (fgets(buf, n + 1, f) == NULL) {
		lg(LOG_ERR, "corsareadtoken: fgets %s: failed", fn);
		(void)fclose(f);
		return (NULL);
	}
	(void)fclose(f);
	cp = buf + n - 1;
	if (cp >= buf && *cp == '\n')
	    *cp = '\0';

	return (buf);
}

/* Return true if the address is an allowable (non-local) filter candidate */
static int
goodfilter(struct addr *addrp)
{
	int i;
	struct addr *a;
	struct array *dp;

	dp = &cf->c_corsalocalnets_array;
	for (i = 0, a = cf->c_corsalocalnets; i < dp->len; ++i, ++a)
		if (insubnet(a, addrp))
			return (0);

	return (1);
}
