#!/usr/local/bin/python
# @(#) $Id: acld.py 1388 2018-11-15 23:54:22Z leres $ (LBL)
"""Communicate with acld"""

from __future__ import print_function

import errno
import os
import random
import select
import socket
import time
import types

DEFAULTINFLIGHT = 100
DEFAULTTIMEOUT = 90
DEFAULTADDR = '127.0.0.1'
DEFAULTPORT = 1965

FAILED_SUFFIX = '-failed'

FILE_CMDS = (
    'addwhitelist',
    'drop',
    'filter',
    'nofilter',
    'nonullzero',
    'nullzero',
    'query',
    'queryfilter',
    'querynullzero',
    'querywhitelist',
    'remwhitelist',
    'restore',
)

CHECK_MAP = {
    'filter': 'queryfilter',
    'nullzero': 'querynullzero',
}

DRYRUN_MAP = {
    'filter': 'queryfilter',
    'nofilter': 'queryfilter',
    'nonullzero': 'querynullzero',
    'nullzero': 'querynullzero',
}

SINGLE_CMDS = (
    'state',
    'listnullzero',
    'listroute',
    'whitelist',
)

DOUBLE_CMDS = (
    'compact',
    'listacl',
    'listprefix',
)

TRIPLE_CMDS = (
    'blockhosthost',
    'restorehosthost',
    'dropudpport',
    'restoreudpport',
    'droptcpport',
    'restoretcpport',
    'droptcpsynport',
    'restoretcpsynport',
    'permitudpdsthostport',
    'unpermitudpdsthostport',
    'permittcpdsthostport',
    'unpermittcpdsthostport',
    'prefix',
    'noprefix',
)

READONLY_CMDS = (
    'listacl',
    'listnullzero',
    'listroute',
    'queryfilter',
    'query',
    'querynullzero',
    'querywhitelist',
    'state',
    'whitelist',
)

class acld(object):
    """Communicate with acld"""
    def __init__(self, check=False, user=None, host=None, debug=False,
	dryrun=False, inflight=DEFAULTINFLIGHT, timeout=DEFAULTTIMEOUT,
	debugfile=None):
	"""Create an acld instance"""
	self.ibuf = ''
	if not user:
	    self.user = os.getlogin()
	else:
	    self.user = user
	if not host:
	    self.host = socket.gethostname()
	else:
	    self.host = host
	self.ident = '{%s@%s}' % (self.user, self.host)
	self.check = check
	self.debug = debug
	self.dryrun = dryrun
	self.inflight = inflight
	self.timeout = timeout
	self.debugfile = debugfile

	self.s = None
	self.cookie = 0
	self.inflightrequests = []

    def close(self):
	"""Close the acld socket"""
	if self.s:
	    # Send the exit
	    scookie = str(self.cookie)
	    self.cookie += 1
	    cmds = ['exit', scookie]
	    cmdline = ' '.join(cmds)
	    tup = [cmdline]
	    self.sendline(tup)
	    self.inflightrequests.append(cmds[0:2])
	    _tret, _last_tup, _payload = self.getresponse()

	    # Close the socket
	    self.s.close()
	    self.s = None

    def getacldline(self):
	"""Get a line from the acld socket and an optional error message"""
	while True:
	    tup = self.ibuf.split('\r\n', 1)
	    if len(tup) > 1:
		self.ibuf = tup[1]
		return (tup[0], None)

	    # Read more
	    rlist = [self.s]
	    wlist = []
	    xlist = []
	    timeout = self.timeout
	    try:
		rlist, wlist, xlist = select.select(rlist, wlist, xlist,
		    timeout)
	    except socket.error as e:
		errmsg = 'getacldline: select: %s' % e.strerror
		return (None, errmsg)

	    # Check for timeout
	    if not rlist and not wlist and not xlist:
		# XXX
		return (None,
		    'getacldline: acld read timeout (%.1f seconds)' % timeout)

	    if self.s in rlist:
		buf = self.s.recv(1024)
		cc = len(buf)
		if cc > 0:
		    # XXX
		    if self.debugfile:
			for line in buf.split('\r\n'):
			    print('<= %s' % line, file=self.debugfile)
		    self.ibuf += buf

    def getresponse(self):
	"""Reap response, return status, last tup and payload/error message"""
	ret = 0
	line, errmsg = self.getacldline()
	if errmsg:
	    return (1, [], errmsg)

	tup = line.split()
	n = len(tup)

	if n < 3 or n > 4:
	    errmsg = 'getresponse: Invalid response: %s' % line
	    return (1, [], errmsg)
	cookie = tup[1]

	if not self.inflightrequests:
	    return (1, [], 'Missing inflightrequests!')

	if cookie == self.inflightrequests[0][1]:
	    last_tup = self.inflightrequests.pop(0)
	else:
	    # Find the index of the request with this cookie
	    try:
		i = [x[1] for x in self.inflightrequests].index(cookie)
		last_tup = self.inflightrequests.pop(i)
	    except ValueError:
		last_tup = []

	    if not last_tup:
		errmsg = 'getresponse: Unexpected cookie: %s' % cookie
		if n == 4:
		    while True:
			line, errmsg2 = self.getacldline()
			if errmsg2:
			    errmsg = '\n'.join([errmsg, errmsg2])
			if line == '.':
			    break
		return (1, [], errmsg)

	what = last_tup[0]
	extended_what = tup[2]

	# Read optional payload
	payloads = []
	if n == 4:
	    if tup[3] != '-':
		errmsg = 'getresponse: Missing continuation: %s' % line
		return (1, [], errmsg)

	    while True:
		line, errmsg = self.getacldline()
		if line == '.':
		    break
		payloads.append(line)
		if errmsg:
		    payloads.append(line)

	if extended_what.endswith(FAILED_SUFFIX):
	    extended_what = extended_what[:-len(FAILED_SUFFIX)]
	    ret = 1

	# Don't check what against extended_what, latter might be "unknown"

	if payloads:
	    payload = '\n'.join(payloads)
	else:
	    payload = None

	return (ret, last_tup, payload)

    def makeconnection(self, servers):
	"""Takes list of server address and port pairs and returns
	   a socket of the first server to connect and an optional
	   errno and error message"""
	sockets = []
	idents = []
	errmsgs = []
	errnums = []
	for i, server in enumerate(servers):
	    addr, port = server
	    ident = '%s.%d' % (addr, port)
	    if '.' in addr:
		fam = socket.AF_INET
	    elif ':' in addr:
		fam = socket.AF_INET6
	    else:
		raise ValueError('Invalid ip address: %s' % addr)
	    s = socket.socket(fam, socket.SOCK_STREAM)
	    # Turn non-blocking off for connect
	    s.setblocking(0)
	    code = s.connect_ex(tuple(server))
	    s.setblocking(1)

	    if code != 0 and code != errno.EINPROGRESS:
		errmsg = '%s: %s' % (ident, os.strerror(code))
		errmsgs.append(errmsg)
		errnums.append(code)
		s.close()
		continue

	    sockets.append(s)
	    idents.append(ident)


	# Loop on select() until we have success or timeout
	ident = None
	s = None
	timeout = float(self.timeout)
	while not s and timeout > 0.0 and sockets:
	    rlist = sockets[:]
	    rlist, _wlist, _xlist = select.select(rlist, [], [], timeout)
	    if not rlist:
		errmsgs.append('timeout')
		errnums.append(errno.ETIMEDOUT)
		continue

	    # Check for connect errors, remove failed candidates
	    for i, s2 in enumerate(rlist):
		try:
		    s2.recv(0)
		except socket.error as e:
		    rlist[i] = None
		    j = sockets.index(s2)
		    ident = idents[j]
		    errmsg = '%s: %s' % (ident, e.strerror)
		    errmsgs.append(errmsg)
		    errnums.append(e.errno)
		    sockets[j].close()
		    sockets.pop(j)
		    idents.pop(j)
		    continue

	    # Lose the losers
	    rlist = filter(None, rlist)

	    if rlist:
		# Pick a winner, get the ident and remove from socket from list
		s = random.choice(rlist)
		i = sockets.index(s)
		ident = idents.pop(i)
		sockets.pop(i)

	# Clean up
	for s2 in sockets:
	    s2.close()

	# If we succeeded just ignore errors accumulated along the way
	if s:
	    return (s, 0, None)

	# Not sure if this can happen
	if not errmsgs:
	    errmsgs.append('failed to open any')
	    errnums.append(errno.EDOM)

	# Return the lowest errno collected and all of the error messages
	errnums.sort()
	return (None, errnums[0], ', '.join(errmsgs))

    def open(self, addr=None, port=None, servers=None):
	"""Open a socket to acld, return an optional error message
	   With no arguments, opens DEFAULTADDR/DEFAULTPORT
	   With addr or port argument, fill in from defaults
	   With servers argument only first addr/port pair to connect"""
	# Forwards compatibility allowing servers as only argument
	if isinstance(addr, types.ListType):
	    if port:
		raise ValueError('addr can not be a list')
	    if not servers:
		servers = addr
		addr = None

	if addr or port:
	    # Disallow addr/port and servers
	    if servers:
		raise ValueError('Cannot specify addr or port and servers')

	    # Fill in ports for single server connect
	    if not addr:
		addr = DEFAULTADDR
	    if not port:
		port = DEFAULTPORT
	    servers = [[addr, port]]
	elif not servers:
	    # No args backwards compatibility
	    servers = [[addr, port]]

	tries = 3
	delay = 0.5
	errmsg = None
	while tries > 0:
	    errmsg = None
	    errnum = -1
	    tries -= 1

	    # Try to make a connection
	    self.s, errnum, errmsg = self.makeconnection(servers)

	    # Retry for some errors
	    # EPERM means the source port was in use
	    if errmsg and errnum in (errno.ECONNREFUSED, errno.EPERM):
		if tries > 0:
		    time.sleep(delay)
		    delay = delay * 2
		    continue
		return errmsg
	    break

	if errmsg:
	    return errmsg
	self.s.setblocking(0)

	self.cookie = 0
	self.inflightrequests = [['acld', '0']]
	ret, _last_tup, payload = self.getresponse()
	if ret:
	    errmsgs = ['bad greeting from acld']
	    if payload:
		errmsgs.append(payload)
	    self.s.close()
	    self.s = None
	    return ': '.join(errmsgs)

	return None

    def process(self, command, items, last=None, comments=None):
	"""Process list of items (addresses or ports)
	   If we're checking, always issue the check command first
	   When we get a check result, optionally issue the real command
	   Return a return status, a list of messages and an optional
	   list of error messages
	   Optionally append "last" to each command to allow multiple
	   port blocks and host port exceptions per commit"""
	msgs = []
	errmsgs = []
	self.inflightrequests = []
	ret = 0

	# Handle check and dryrun modes
	detail = '?'
	if self.dryrun:
	    command2 = DRYRUN_MAP.get(command)
	    detail = 'dry run'
	elif self.check:
	    command2 = CHECK_MAP.get(command)
	    detail = 'check'
	else:
	    command2 = None

	if (self.dryrun or self.check) and not command2:
	    errmsg = "Can't find %s command for %s" % (detail, command)
	    errmsgs.append(errmsg)
	    return (1, msgs, errmsgs)

	# Loop on items from the list
	self.cookie = 0
	for item in items:
	    scookie = str(self.cookie)
	    self.cookie += 1

	    # Build the next command
	    if command2:
		cmds = [command2, scookie, item]
		if last:
		    cmds.append(last)
		cmdline = ' '.join(cmds)
		tup = [cmdline]
	    else:
		cmds = [command, scookie, item]
		if last:
		    cmds.append(last)
		cmds.append('-')
		cmdline = ' '.join(cmds)
		tup = [cmdline, self.ident]
		if comments:
		    tup.extend(comments)
		tup.append('.')

	    # Send the command
	    self.sendline(tup)
	    self.inflightrequests.append(cmds[0:3])

	    # Reap
	    if len(self.inflightrequests) >= self.inflight:
		tret, last_tup, payload = self.getresponse()

		# Is this a check result?
		# XXX
		# What about filter?
		# XXX
		if command2 and last_tup[0] == command2:
		    if self.dryrun:
			if payload and payload.startswith('N '):
			    fmt = 'Note: %s is already nullzero routed'
			    msg = fmt % last_tup[2]
			    msgs.append(msg)
			continue

		    if tret:
			scookie = str(self.cookie)
			self.cookie += 1
			item = last_tup[2]
			cmds = [command, scookie, item, '-']
			cmdline = ' '.join(cmds)

			# Send the command
			tup = [cmdline, self.ident]
			if comments:
			    tup.extend(comments)
			tup.append('.')
			self.sendline(tup)
			self.inflightrequests.append(cmds[0:3])
		    continue

		# Command result
		if tret:
		    ret = 1
		    # Append item
		    if len(last_tup) > 2:
			payload = ' '.join([payload, last_tup[2]])
		if payload:
		    msgs.append(payload)

	# Reap the stragglers
	while len(self.inflightrequests) > 0:
	    tret, last_tup, payload = self.getresponse()

	    # Is this a check result?
	    if command2 and last_tup[0] == command2:
		if self.dryrun:
		    # What about filter?
		    # XXX
		    if payload and payload.startswith('N '):
			msg = ('Note: %s is already nullzero routed' %
			    last_tup[2])
			msgs.append(msg)
		    continue

		if tret:
		    scookie = str(self.cookie)
		    self.cookie += 1
		    item = last_tup[2]
		    cmds = [command, scookie, item, '-']
		    cmdline = ' '.join(cmds)

		    # Send the command
		    tup = [cmdline, self.ident]
		    if comments:
			tup.extend(comments)
		    tup.append('.')
		    self.sendline(tup)
		    self.inflightrequests.append(cmds[0:3])
		continue

	    # Command result
	    if tret:
		ret = 1
		# Append item
		if len(last_tup) > 2:
		    payload = ' '.join([payload, last_tup[2]])
	    if payload:
		msgs.append(payload)

	return (ret, msgs, errmsgs)

    def processone(self, args, comments=None):
	"""Process a single command
	   Return a return status, a list of messages and an optional
	   list of error messages"""
	msgs = []
	errmsgs = []
	self.cookie = 0
	ret = 0
	command = args[0]

	# Handle check and dryrun modes
	detail = '?'
	if self.dryrun:
	    command2 = DRYRUN_MAP.get(command)
	    detail = 'dryrun'
	elif self.check:
	    command2 = CHECK_MAP.get(command)
	    detail = 'check'
	else:
	    command2 = None

	if (self.dryrun or self.check) and not command2:
	    errmsg = "Can't find %s command for %s" % (detail, command)
	    errmsgs.append(errmsg)
	    return (1, msgs, errmsgs)

	# Run the check command
	if command2:
	    # Build the command
	    scookie = str(self.cookie)
	    self.cookie += 1
	    cmds = [command2, scookie]
	    cmds += args[1:]
	    cmdline = ' '.join(cmds)

	    # Send the command
	    tup = [cmdline]
	    self.sendline(tup)
	    self.inflightrequests = [cmds]
	    tret, last_tup, payload = self.getresponse()
	    if self.dryrun:
		if command in ('nofilter', 'nonullzero'):
		    # Return same queryfilter/querynullzero result
		    return (tret, msgs, errmsgs)

		# filter/nullzero cases
		if tret == 0:
		    # report the filter/nullzero already exists
		    return (2, msgs, errmsgs)
		else:
		    # pretend we added a filter/nullzero
		    return (0, msgs, errmsgs)

	    # Normal case
	    if tret == 0:
		msg = 'Skipping %s: (%s is yes)' % (' '.join(args), command)
		msgs.append(msg)
		return (2, msgs, errmsgs)

	# Build the command
	cmds = args[:]
	self.inflightrequests = [cmds]
	scookie = str(self.cookie)
	self.cookie += 1
	cmds.insert(1, scookie)
	cmds.append('-')
	cmdline = ' '.join(cmds)

	if self.dryrun:
	    msg = '+ %s' % cmdline
	    msgs.append(msg)
	else:
	    # Send the command
	    tup = [cmdline, self.ident]
	    if comments:
		tup.extend(comments)
	    tup.append('.')
	    self.sendline(tup)
	    ret, last_tup, payload = self.getresponse()
	    if payload:
		if ret and len(last_tup) > 2 and last_tup[2] not in payload:
		    payload = ' '.join([payload, last_tup[2]])
		msgs.append(payload)
		# Return the special ""already exists" code
		if 'already' in payload:
		    ret = 2

	return (ret, msgs, errmsgs)

    def sendline(self, tup):
	"""Add \r\n's to each string in the list and concatenate
	   into a string and deal with non blocking I/O"""
	if self.debugfile:
	    for line in tup:
		print('=> %s' % line, file=self.debugfile)
	buf = ''.join([x + '\r\n' for x in tup])
	while True:
	    rlist = []
	    wlist = [self.s]
	    xlist = []
	    timeout = 1.0
	    rlist, wlist, xlist = select.select(rlist, wlist, xlist, timeout)

	    if self.s in wlist:
		self.s.send(buf)
		break
