/*
 * Copyright (c) 2006, 2007, 2008, 2009, 2010, 2011, 2013, 2015, 2018
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: whitelist.c 1414 2018-11-21 06:19:57Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <time.h>
#include <unistd.h>

#include "acld.h"
#include "child.h"

/* Forwards */
static void whiteliststradd(const char *, int);

/* Returns addr if address or network is whitelisted else NULL */
struct addr *
whitelist(struct addr *addr)
{
	int i, w;
	struct addr *a;
	struct array *dp;

	dp = &cf->c_whitelist_array;
	if (dp->len == 0)
		return (NULL);

	/* Return on any overlap */
	w = maskwidth(addr);
	for (i = 0, a = cf->c_whitelist; i < dp->len; ++i, ++a) {
		if (a->family != addr->family)
			continue;
		if (w >= maskwidth(a)) {
			/* Test addr is more specific than whitelist addr */
			if (insubnet(a, addr))
				return (a);
		} else {
			/* Test addr is less specific than whitelist addr */
			if (insubnet(addr, a))
				return (a);
		}
	}

	return (NULL);
}

/* Does not check to see if the addr is already on the list */
void
whitelistadd(struct addr *addr)
{
	struct addr *a;
	struct array *dp;

	dp = &cf->c_whitelist_array;
	DYNARRAY(dp, cf->c_whitelist, dp->len + 1, "whitelistadd");
	a = cf->c_whitelist + dp->len;
	*a = *addr;
	++dp->len;
}

int
whitelistdump(struct iobuf *ip)
{
	int i;
	struct addr *a;
	struct array *dp;

	dp = &cf->c_whitelist_array;
	for (i = 0, a = cf->c_whitelist; i < dp->len; ++i, ++a)
		ioappendfmt(ip, "%s", addr2str(a));
	return (dp->len);
}

/* Returns number of matching whitelist entries reported */
int
whitelistquery(struct addr *addr, struct iobuf *ip)
{
	int i, cnt, w;
	struct addr *a;
	struct array *dp;

	dp = &cf->c_whitelist_array;
	if (dp->len == 0)
		return (0);

	/* Report on any overlap */
	w = maskwidth(addr);
	cnt = 0;
	for (i = 0, a = cf->c_whitelist; i < dp->len; ++i, ++a) {
		if (a->family != addr->family)
			continue;
		if (w >= maskwidth(a)) {
			/* Test addr is more specific than whitelist addr */
			if (insubnet(a, addr)) {
				ioappendfmt(ip, "%s", addr2str(a));
				++cnt;
			}
		} else {
			/* Test addr is less specific than whitelist addr */
			if (insubnet(addr, a)) {
				ioappendfmt(ip, "%s", addr2str(a));
				++cnt;
			}
		}
	}

	return (cnt);
}

void
whitelistread()
{
	int n;
	char *cp, *cp2;
	const char *errstr;
	FILE *f;
	char buf[1024];

	if (cf->c_whitelistfn == NULL)
		return;

	f = fopen(cf->c_whitelistfn, "r");
	if (f == NULL) {
		if (errno != ENOENT) {
			lg(LOG_ERR, "whitelistread: fopen %s: %s",
			    cf->c_whitelistfn, strerror(errno));
			exit(1);
		}
		lg(LOG_INFO,
		    "whitelistread: %s missing; creating empty whitelist file",
		    cf->c_whitelistfn);
		errstr = whitelistupdate();
		if (errstr != NULL) {
			lg(LOG_ERR, "whitelistread: whitelistupdate: %s",
			    errstr);
			exit(1);
		}
		return;
	}

	n = 0;
	while (fgets(buf, sizeof(buf), f) != NULL) {
		++n;

		/* Eat comments */
		cp = strchr(buf, '#');
		if (cp)
			*cp = '\0';

		/* Eat trailing whitesapce */
		cp = buf + strlen(buf) - 1;
		while (cp >= buf && isspace((int)*cp))
			*cp-- = '\0';
		cp = buf;

		/* Skip blank lines */
		if (*cp == '\n' || *cp == '\0')
			continue;

		/* Get address/network */
		cp2 = cp;
		while (!isspace((int)*cp) && *cp != '\0')
			++cp;
		*cp++ = '\0';

		whiteliststradd(cp2, n);
	}
	(void)fclose(f);
}

int
whitelistrem(struct addr *addr)
{
	int i, j;
	struct addr *a;
	struct array *dp;

	dp = &cf->c_whitelist_array;
	if (dp->len == 0)
		return (0);

	for (i = 0, a = cf->c_whitelist; i < dp->len; ++i, ++a)
		if (ADDREQ(a, addr)) {
			j = dp->len - (i + 1);
			if (j > 0)
				memmove(cf->c_whitelist + i,
				    cf->c_whitelist + i + 1, j * sizeof(*a));
			--dp->len;
			ADDRZERO(a + dp->len);
			return (1);
		}

	*a = *addr;
	++dp->len;

	return (0);
}

/* Look for the exact address in the whitelist */
struct addr *
whitelistsearch(struct addr *addr)
{
	int i;
	struct addr *a;
	struct array *dp;

	dp = &cf->c_whitelist_array;
	if (dp->len == 0)
		return (NULL);

	for (i = 0, a = cf->c_whitelist; i < dp->len; ++i, ++a)
		if (ADDREQ(a, addr))
			return (a);

	return (NULL);
}

static void
whiteliststradd(const char *str, int n)
{
	const char *p;
	struct addr *a;
	struct addr addr;
	char detail[1024];

	(void)snprintf(detail, sizeof(detail), "%s:%d", cf->c_whitelistfn, n);

	/* Eat leading dot; done if there was only one */
	if (*str == '.') {
		++str;
		if (*str == '\n'|| *str == '\0') {
			lg(LOG_ERR, "whiteliststradd: %s: junk \"%.80s\"",
			    detail,  str);
			return;
		}
	}

	/* Eat comments */
	if (*str == '#')
		return;

	a = &addr;
	memset(a, 0, sizeof(*a));

	p = extractaddr(str, NULL, a);
	if (p != NULL) {
		lg(LOG_ERR, "whiteliststradd: %s: extractaddr dst %.80s: %s",
		    detail, str, p);
		exit(EX_SOFTWARE);
	}

	if (whitelistsearch(a)) {
		lg(LOG_ERR, "whiteliststradd: %s: duplicate %.80s",
		    detail, str);
		return;
	}
	whitelistadd(a);
}

const char *
whitelistupdate()
{
	int i;
	FILE *f;
	time_t now;
	struct addr *a;
	struct array *dp;
	struct tm *tm;
	char date[64];
	char zone[32];
	char temp[256];
	static char errstr[256];
#ifdef HAVE_ALTZONE
	char ch;
	int z;
#endif
	if (cf->c_whitelistfn == NULL)
		return ("whitelistupdate: whitelist file not configured");

	(void)snprintf(temp, sizeof(temp), "%.*s-",
	    (int)sizeof(temp) - 2, cf->c_whitelistfn);
	f = fopen(temp, "w");
	if (f == NULL) {
		(void)snprintf(errstr, sizeof(errstr),
		    "whitelistwrite: fopen %s: %s", temp, strerror(errno));
		lg(LOG_ERR, "%s", errstr);
		return (errstr);
	}
	fprintf(f, "# WARNING: dynamically updated; do not edit\n");
	now = time(NULL);
	tm = localtime(&now);
	(void)strftime(date, sizeof(date), "%d-%b-%Y %T", tm);
#ifndef HAVE_ALTZONE
	(void)strftime(zone, sizeof(zone), "%z", tm);
#else
	z = altzone / -36;
	ch = '+';
	if (z < 0) {
		z = -z;
		ch = '-';
	}
	(void)snprintf(zone, sizeof(zone), "%c%04d", ch, z);
#endif
	fprintf(f, "# Last update: %s %s\n", date, zone);

	dp = &cf->c_whitelist_array;
	for (i = 0, a = cf->c_whitelist; i < dp->len; ++i, ++a)
		fprintf(f, "%s\n", addr2str(a));

	if (fclose(f) != 0) {
		(void)snprintf(errstr, sizeof(errstr),
		    "whitelistwrite: fclose %s: %s", temp, strerror(errno));
		lg(LOG_ERR, "%s", errstr);
		return (errstr);
	}
	if (rename(temp, cf->c_whitelistfn) < 0) {
		(void)snprintf(errstr, sizeof(errstr),
		    "whitelistwrite: rename %s -> %s: %s",
		    temp, cf->c_whitelistfn, strerror(errno));
		lg(LOG_ERR, "%s", errstr);
		return (errstr);
	}
	return (NULL);
}
